import argparse
import json
import os
import shutil
import time
from typing import Dict

import pandas as pd

from ..preprocessing import preprocessor
from ..serving.predictor import BanditPredictor
from ..utils import feature_importance, model_constructors, model_trainers, utils

logger = utils.get_logger(__name__)


def num_float_dim(data):
    return len(data["X_float"].columns)


def train(
    training_df: pd.DataFrame,
    ml_config: Dict,
    feature_config: Dict,
    predictor_save_dir: str = None,
) -> BanditPredictor:

    start = time.time()

    utils.validate_ml_config(ml_config)

    logger.info("Checking schema of training data...")
    utils.validate_training_data_schema(training_df)

    if len(training_df) == 0:
        logger.error(f"Got no raws of training data. Training aborted.")
        return None

    logger.info(f"Got {len(training_df)} rows of training data.")
    logger.info(training_df.head())

    utils.fancy_print("Kicking off data preprocessing")

    # always add decision as a feature to use if not using all features
    features_to_use = ml_config["features"].get("features_to_use", ["*"])
    if features_to_use != ["*"]:
        features_to_use.append(preprocessor.DECISION_FEATURE_NAME)
    features_to_use = list(set(features_to_use))
    dense_features_to_use = ml_config["features"].get("dense_features_to_use", ["*"])

    data = preprocessor.preprocess_data(
        training_df,
        feature_config,
        ml_config["reward_type"],
        features_to_use,
        dense_features_to_use,
    )
    X, y = preprocessor.data_to_pytorch(data)

    model_type = ml_config["model_type"]
    model_params = ml_config["model_params"][model_type]
    reward_type = ml_config["reward_type"]

    feature_importance_params = ml_config.get("feature_importance", {})
    if feature_importance_params.get("calc_feature_importance", False):
        # calculate feature importances - only works on non id list features at this time
        utils.fancy_print("Calculating feature importances")
        feature_scores = feature_importance.calculate_feature_importance(
            reward_type=reward_type,
            feature_names=data["final_float_feature_order"],
            X=X,
            y=y,
        )
        feature_importance.display_feature_importances(feature_scores)

        # TODO: Make keeping the top "n" features work in predictor. Right now
        # using this feature breaks predictor, so don't use it in a final model,
        # just use it to experiment in seeing how model performance is.
        if feature_importance_params.get("keep_only_top_n", False):
            utils.fancy_print("Keeping only top N features")
            X, final_float_feature_order = feature_importance.keep_top_n_features(
                n=feature_importance_params["n"],
                X=X,
                feature_order=data["final_float_feature_order"],
                feature_scores=feature_scores,
            )
            data["final_float_feature_order"] = final_float_feature_order
            logger.info(f"Keeping top {feature_importance_params['n']} features:")
            logger.info(final_float_feature_order)

    utils.fancy_print("Starting training")
    # build the model
    if model_type == "neural_bandit":
        model_spec, model = model_constructors.build_pytorch_net(
            feature_specs=feature_config["features"],
            product_sets=feature_config["product_sets"],
            float_feature_order=data["final_float_feature_order"],
            id_feature_order=data["final_id_feature_order"],
            reward_type=reward_type,
            layers=model_params["layers"],
            activations=model_params["activations"],
            dropout_ratio=model_params["dropout_ratio"],
            input_dim=num_float_dim(data),
        )
        logger.info(f"Initialized model: {model}")
    elif model_type == "linear_bandit":
        assert utils.pset_features_have_dense(feature_config["features"]), (
            "Linear models require that product set features have associated"
            "dense representations."
        )
        model = model_constructors.build_linear_model(
            reward_type=reward_type,
            penalty=model_params.get("penalty"),
            alpha=model_params.get("alpha"),
        )
        model_spec = None
    elif model_type == "gbdt_bandit":
        assert utils.pset_features_have_dense(feature_config["features"]), (
            "GBDT models require that product set features have associated"
            "dense representations."
        )
        model = model_constructors.build_gbdt(
            reward_type=reward_type,
            learning_rate=model_params["learning_rate"],
            n_estimators=model_params["n_estimators"],
            max_depth=model_params["max_depth"],
        )
        model_spec = None
    elif model_type == "random_forest_bandit":
        assert utils.pset_features_have_dense(feature_config["features"]), (
            "Random forest models require that product set features have associated"
            "dense representations."
        )
        model = model_constructors.build_random_forest(
            reward_type=reward_type,
            n_estimators=model_params["n_estimators"],
            max_depth=model_params["max_depth"],
        )
        model_spec = None

    # build the predictor
    predictor = BanditPredictor(
        feature_config=feature_config,
        float_feature_order=data["float_feature_order"],
        id_feature_order=data["id_feature_order"],
        id_feature_str_to_int_map=data["id_feature_str_to_int_map"],
        transforms=data["transforms"],
        imputers=data["imputers"],
        model=model,
        model_type=model_type,
        reward_type=reward_type,
        model_spec=model_spec,
        dense_features_to_use=dense_features_to_use,
    )

    # train the model
    if model_type == "neural_bandit":
        logger.info(f"Training {model_type} for {model_params['max_epochs']} epochs")
        skorch_net = model_trainers.fit_custom_pytorch_module_w_skorch(
            reward_type=reward_type,
            model=predictor.model,
            X=X,
            y=y,
            hyperparams=model_params,
            train_percent=ml_config["train_percent"],
        )
    elif model_type in ("gbdt_bandit", "random_forest_bandit", "linear_bandit"):
        logger.info(f"Training {model_type}")
        sklearn_model, _ = model_trainers.fit_sklearn_model(
            reward_type=reward_type,
            model=model,
            X=X,
            y=y,
            train_percent=ml_config["train_percent"],
        )

    if predictor_save_dir is not None:
        logger.info("Saving predictor artifacts to disk...")
        model_name = ml_config.get("model_name", "model")

        save_dir = f"{predictor_save_dir}/{model_name}/"
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)

        predictor_net_path = f"{save_dir}/{model_name}.pt"
        predictor_config_path = f"{save_dir}/{model_name}.json"
        predictor.config_to_file(predictor_config_path)
        predictor.model_to_file(predictor_net_path)
        shutil.make_archive(save_dir, "zip", save_dir)

    logger.info(f"Traning took {time.time() - start} seconds.")

    return predictor



import json
import logging
import math
from typing import Dict, NoReturn

import pandas as pd

VALID_MODEL_TYPES = (
    "linear_bandit",
    "neural_bandit",
    "random_forest_bandit",
    "gbdt_bandit",
    "dqn",
    "qr_dqn",
)
VALID_REWARD_TYPES = ("regression", "binary")


def read_config(config_path):
    with open(config_path) as json_file:
        config = json.load(json_file)
    return config


def get_logger(name: str):
    logging.basicConfig(
        format="[%(asctime)s %(levelname)-3s] %(message)s",
        level=logging.INFO,
        datefmt="%H:%M:%S",
    )
    return logging.getLogger(name)


def fancy_print(text: str, color="blue", size=60):
    ansi_color = "\033[94m"
    if color == "green":
        ansi_color = "\033[95m"
    elif color == "blue":
        ansi_color = "\033[94m"
    else:
        raise Exception(f"Color {color} not supported")

    end_color = "\033[0m"
    str_len = len(text)
    padding = math.ceil((size - str_len) / 2)
    header_len = padding * 2 + str_len + 2
    border = "#" * header_len
    message = "#" * padding + " " + text + " " + "#" * padding
    print(f"{ansi_color}\n{border}\n{message}\n{border}\n{end_color}")


def color_text(text: str, color="blue"):
    ansi_color = "\033[94m"
    if color == "green":
        ansi_color = "\033[95m"
    elif color == "blue":
        ansi_color = "\033[94m"
    else:
        raise Exception(f"Color {color} not supported")

    end_color = "\033[0m"
    return f"{ansi_color}{text}{end_color}"


def validate_ml_config(ml_config: Dict) -> NoReturn:
    assert "model_type" in ml_config
    assert "model_params" in ml_config
    assert "reward_type" in ml_config

    model_type = ml_config["model_type"]
    model_params = ml_config["model_params"]
    reward_type = ml_config["reward_type"]

    assert (
        model_type in VALID_MODEL_TYPES
    ), f"Model type {model_type} not supported. Valid model types are {VALID_MODEL_TYPES}"
    assert model_type in model_params
    assert (
        reward_type in VALID_REWARD_TYPES
    ), f"Reward type {reward_type} not supported. Valid reward types are {VALID_REWARD_TYPES}"


def validate_training_data_schema(training_df: pd.DataFrame) -> NoReturn:
    for col in ["context", "decision", "reward", "mdp_id", "sequence_number"]:
        assert col in training_df.columns


def pset_features_have_dense(features: Dict) -> NoReturn:
    for feature, meta in features.items():
        if meta["type"] == "P":
            if not meta["use_dense"]:
                return False
    return True



from __future__ import unicode_literals

import json
import numbers
import time
from typing import Dict, List, Optional

from google.cloud import bigquery

from .bq import Table
from .v1 import Decision, Reward

FEEDBACK_TABLE = "feedback"
REWARD_TYPE_IMMEDIATE = "immediate"
REWARD_TYPE_DELAYED = "delayed"


SCHEMA = [
    bigquery.SchemaField("timestamp", "TIMESTAMP", mode="REQUIRED"),
    bigquery.SchemaField("company", "STRING", mode="NULLABLE"),
    bigquery.SchemaField("experiment_id", "STRING", mode="NULLABLE"),
    bigquery.SchemaField("mdp_id", "STRING", mode="NULLABLE"),
    bigquery.SchemaField("type", "STRING", mode="NULLABLE"),
    bigquery.SchemaField("choice_id", "STRING", mode="NULLABLE"),
    bigquery.SchemaField("choice_score", "FLOAT64", mode="NULLABLE"),
    bigquery.SchemaField("variant_id", "INT64", mode="NULLABLE"),
    bigquery.SchemaField("variant_slug", "STRING", mode="NULLABLE"),
    bigquery.SchemaField("decision_id", "STRING", mode="NULLABLE"),
    bigquery.SchemaField("decision_context_json", "STRING", mode="NULLABLE"),
    bigquery.SchemaField("reward_type", "STRING", mode="NULLABLE"),
    bigquery.SchemaField("reward_metric", "STRING", mode="NULLABLE"),
    bigquery.SchemaField("reward_value", "FLOAT64", mode="NULLABLE"),
]


class Feedback:
    def __init__(self, **kwargs):
        self.timestamp: int = kwargs.get("timestamp", time.time())
        self.company: str = kwargs.get("company", None)
        self.experiment_id: str = kwargs.get("experiment_id", None)
        self.mdp_id: str = kwargs.get("mdp_id", None)
        self.type: str = kwargs.get("type", None)
        self.choice_id: Optional[str] = kwargs.get("choice_id", None)
        self.choice_score: Optional[float] = kwargs.get("choice_score", None)
        self.variant_id: Optional[int] = kwargs.get("variant_id", None)
        self.variant_slug: Optional[str] = kwargs.get("variant_slug", None)
        self.decision_id: Optional[str] = kwargs.get("decision_id", None)
        self.decision_context_json: Optional[str] = kwargs.get(
            "decision_context_json", None
        )
        self.reward_type: Optional[str] = kwargs.get("reward_type", None)
        self.reward_metric: Optional[str] = kwargs.get("reward_metric", None)
        self.reward_value: Optional[float] = kwargs.get("reward_value", None)

    def to_dict(self) -> Dict:
        return self.__dict__

    @classmethod
    def from_decision(cls, company: str, decision: Decision) -> "Feedback":
        """
        Maps a v1 `decision` record to a v2 `feedback` record.
        `variant_slug` is unmappable without API access (OSS) or DB access (internal).
        """
        return Feedback(
            timestamp=decision.get("ts", None),
            company=company,
            type="decision",
            choice_id=decision.get("decision", None),
            choice_score=decision.get("score", None),
            decision_context_json=decision.get("context", None),
            variant_id=decision.get("variation_id", None),
            **decision,
        )

    @classmethod
    def from_reward(cls, company: str, reward: Reward) -> List["Feedback"]:
        """
        Maps a legacy `reward` record to multiple `feedback` records.
        """

        def create_feedback(reward_key, reward_value) -> "Feedback":
            # uses `choice_id` and `reward_type` from scope
            return Feedback(
                timestamp=reward.get("ts", None),
                company=company,
                type="reward",
                choice_id=choice_id,
                reward_type=reward_type,
                reward_metric=reward_key,
                reward_value=reward_value,
                **reward,
            )

        feedbacks = []
        reward_type = REWARD_TYPE_IMMEDIATE
        if reward.get("decision") is None and reward.get("decision_id") is None:
            reward_type = REWARD_TYPE_DELAYED
        metrics = json.loads(reward["metrics"])
        for metric_key, metric_value in metrics.items():
            if reward_type == REWARD_TYPE_DELAYED:
                assert isinstance(metric_key, str)
                assert isinstance(metric_value, dict)
                choice_id = metric_key
                inner_metrics = metric_value
                for inner_metric_key, inner_metric_value in inner_metrics.items():
                    assert isinstance(inner_metric_key, str)
                    if not isinstance(inner_metric_value, numbers.Number):
                        continue
                    feedbacks.append(
                        create_feedback(inner_metric_key, inner_metric_value)
                    )
            else:
                assert isinstance(metric_key, str)
                if not isinstance(metric_value, numbers.Number):
                    continue
                choice_id = reward.get("decision")
                feedbacks.append(create_feedback(metric_key, metric_value))
        return feedbacks


class FeedbackTable(Table):
    name = FEEDBACK_TABLE
    model = Feedback

    @classmethod
    def create(
        cls, client: bigquery.Client, project_id: str, dataset_id: str, **kwargs
    ) -> "FeedbackTable":
        return super().create(
            client, project_id, dataset_id, schema=SCHEMA, partition=True
        )

    def mapper(self, bq_row) -> Feedback:
        return Feedback(**{k: v for k, v in bq_row.items()})



from typing_extensions import TypedDict

from .bq import Table

DECISIONS_TABLE = "decisions"
REWARDS_TABLE = "rewards"


class Base(TypedDict):
    experiment_id: str
    mdp_id: str
    decision_id: str
    decision: str
    ts: int


class Decision(Base):
    variation_id: int
    context: str
    score: float


class Reward(Base):
    metrics: str


class RewardTable(Table):
    name = REWARDS_TABLE
    partition_field = "_PARTITIONTIME"

    def mapper(self, bq_row) -> Reward:
        return {k: v for k, v in bq_row.items()}


class DecisionTable(Table):
    name = DECISIONS_TABLE
    partition_field = "_PARTITIONTIME"

    def mapper(self, bq_row) -> Decision:
        return {k: v for k, v in bq_row.items()}



import numpy as np

import torch
from sklearn.metrics import accuracy_score, log_loss, mean_squared_error, roc_auc_score
from sklearn.model_selection import train_test_split
from skorch import NeuralNetClassifier, NeuralNetRegressor, dataset

from . import utils

logger = utils.get_logger(__name__)


def fit_custom_pytorch_module_w_skorch(
    reward_type, model, X, y, hyperparams, model_name="", train_percent=0.8
):
    """Fit a custom PyTorch module using Skorch."""
    logger.info(f"Model input dimension: {model.layers[0].in_features}")

    if reward_type == "regression":
        skorch_func = NeuralNetRegressor
    else:
        skorch_func = NeuralNetClassifier
        # torch's nll_loss wants 1-dim tensors & long type tensors.
        y = y.long().squeeze()

    if model_name == "mixture_density_network":

        # MDN loss seemed to stabilize around 100 iterations, lmk if I should remove this!
        num_epochs = 2 * hyperparams["max_epochs"]

        class net_func(skorch_func):
            def get_loss(self, y_pred, y_true, X=None, training=False):

                n = y_pred.shape[0] // 2
                mu = y_pred[:n]
                sigma = y_pred[n:]

                dist = torch.distributions.Normal(loc=mu, scale=sigma)
                loss = torch.mean(-dist.log_prob(y_true))

                return loss

    else:
        num_epochs = hyperparams["max_epochs"]
        net_func = skorch_func

    skorch_net = net_func(
        module=model,
        optimizer=torch.optim.Adam,
        lr=hyperparams["learning_rate"],
        optimizer__weight_decay=hyperparams["l2_decay"],
        max_epochs=num_epochs,
        batch_size=hyperparams["batch_size"],
        iterator_train__shuffle=True,
        train_split=dataset.CVSplit(1 - train_percent),
    )

    skorch_net.fit(X, y)
    return skorch_net


def fit_sklearn_model(reward_type, model, X, y, train_percent=0.8):
    X = X["X_float"]
    logger.info(f"Model input dimension: {X.shape[1]}")

    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=(1 - train_percent)
    )
    model.fit(X_train, y_train)
    training_stats = {}

    if reward_type == "regression":
        mse_train = mean_squared_error(y_train, model.predict(X_train))
        mse_test = mean_squared_error(y_test, model.predict(X_test))
        logger.info(utils.color_text(f"Training MSE: {mse_train}", color="blue"))
        logger.info(utils.color_text(f"Test MSE: {mse_test}", color="green"))
        training_stats["mse_train"] = mse_train
        training_stats["mse_test"] = mse_test
    else:

        # first check what picking the majority class would get you in terms
        # of accuracy and auc as a benchmark
        mode, _ = torch.mode(y_test, dim=0)
        naive_acc_test = accuracy_score(y_test, np.repeat(mode, len(y_test)))
        naive_roc_test = roc_auc_score(y_test, np.repeat(mode, len(y_test)))
        logger.info("Naive majority class model:")
        logger.info(f"Test accuracy: {naive_acc_test}")
        logger.info(f"Test ROC AUC: {naive_roc_test}")
        logger.info("-" * 30)

        # Now actually score the trained model
        train_preds = model.predict(X_train)
        test_preds = model.predict(X_test)
        acc_train = accuracy_score(y_train, train_preds)
        acc_test = accuracy_score(y_test, test_preds)
        roc_train = roc_auc_score(y_train, train_preds)
        roc_test = roc_auc_score(y_test, test_preds)
        lloss_train = log_loss(y_train, model.predict_proba(X_train))
        lloss_test = log_loss(y_test, model.predict_proba(X_test))

        logger.info("Trained model:")
        logger.info(utils.color_text(f"Training accuracy: {acc_train}", color="blue"))
        logger.info(utils.color_text(f"Test accuracy: {acc_test}", color="green"))
        logger.info(utils.color_text(f"Training log loss: {lloss_train}", color="blue"))
        logger.info(utils.color_text(f"Test log loss: {lloss_test}", color="green"))
        logger.info(utils.color_text(f"Train ROC AUC: {roc_train}", color="blue"))
        logger.info(utils.color_text(f"Test ROC AUC: {roc_test}", color="green"))

        training_stats["acc_train"] = acc_train
        training_stats["acc_test"] = acc_test
        training_stats["roc_train"] = roc_train
        training_stats["roc_test"] = roc_test
        training_stats["lloss_train"] = lloss_train
        training_stats["lloss_test"] = lloss_test

    return model, training_stats



from sklearn import ensemble, linear_model

from ..models.embed_dnn import EmbedDnn, build_embedding_spec


def build_pytorch_net(
    feature_specs,
    product_sets,
    float_feature_order,
    id_feature_order,
    reward_type,
    layers,
    activations,
    input_dim,
    dropout_ratio=0.0,
    is_mdn=False,
):
    """Build PyTorch model that will be fed into skorch training."""

    is_classification = reward_type == "binary"
    output_dim = 2 if is_classification else 1

    layers[0], layers[-1] = input_dim, output_dim

    # handle changes of model architecture due to embeddings
    first_layer_dim_increase, embedding_info = build_embedding_spec(
        id_feature_order, feature_specs, product_sets
    )
    layers[0] += first_layer_dim_increase

    model_spec = {
        "layers": layers,
        "activations": activations,
        "dropout_ratio": dropout_ratio,
        "feature_specs": feature_specs,
        "product_sets": product_sets,
        "float_feature_order": float_feature_order,
        "id_feature_order": id_feature_order,
        "embedding_info": embedding_info,
        "is_classification": is_classification,
        "is_mdn": is_mdn,
    }
    return model_spec, EmbedDnn(**model_spec)


def build_gbdt(reward_type, learning_rate=0.1, n_estimators=100, max_depth=3):
    is_classification = reward_type == "binary"
    params = {
        "learning_rate": learning_rate,
        "n_estimators": n_estimators,
        "max_depth": max_depth,
    }

    if is_classification:
        gbdt_model = ensemble.GradientBoostingClassifier(**params)
    else:
        gbdt_model = ensemble.GradientBoostingRegressor(**params)

    return gbdt_model


def build_random_forest(reward_type, n_estimators=100, max_depth=None):
    is_classification = reward_type == "binary"
    params = {"n_estimators": n_estimators, "max_depth": max_depth}

    if is_classification:
        gbdt_model = ensemble.RandomForestClassifier(**params)
    else:
        gbdt_model = ensemble.RandomForestRegressor(**params)

    return gbdt_model


def build_linear_model(reward_type, penalty="l2", alpha=1.0):
    is_classification = reward_type == "binary"
    if is_classification:
        params = {"penalty": penalty}
        linear_model_ = linear_model.LogisticRegression(**params)
    else:
        params = {"alpha": alpha}
        linear_model_ = linear_model.Ridge(**params)

    return linear_model_



"""BQ contains base classes and utilities for working with BigQuery as a BanditML data source."""

from typing import Any, ClassVar, Dict, Iterator, List, Sequence

import google
from google.cloud import bigquery

from typing_extensions import TypedDict


class BQError(TypedDict):
    index: int
    errors: List[str]


class Table:
    """A base class for OO BQ table interactions.

    Subclasses must implement `mapper`, and may override class attribute `name` with the table_id, and
    `partition_field`.
    """

    DEFAULT_BUFFER_SIZE = 10000
    SCHEMA: ClassVar[List[bigquery.SchemaField]] = None

    name: ClassVar[str] = None
    partition_field: ClassVar[str] = "timestamp"

    bq_table: bigquery.Table

    @classmethod
    def create(
        cls,
        client: bigquery.Client,
        project_id: str,
        dataset_id: str,
        schema: List[bigquery.SchemaField] = None,
        partition: bool = False,
    ):
        if cls.name is None:
            raise NotImplementedError(
                "class attribute `name` must be set in order to create table"
            )
        bq_table = bigquery.Table(
            _fulltable(project_id, dataset_id, cls.name),
            schema=schema,
        )
        bq_table.time_partitioning = bigquery.TimePartitioning(
            type_=bigquery.TimePartitioningType.DAY,
            field=cls.partition_field,
        )
        bq_table = client.create_table(bq_table)

        return cls(client, bq_table=bq_table)

    def __init__(
        self,
        client: bigquery.Client,
        bq_table: bigquery.Table = None,
        project_id: str = None,
        dataset_id: str = None,
    ):
        full_path_ok: bool = (
            project_id is not None and dataset_id is not None and self.name is not None
        )
        if bq_table is None and not full_path_ok:
            raise ValueError(
                "bq_table or project_id, dataset_id, and class variable 'name' are required arguments"
            )
        self.client: bigquery.Client = client
        if bq_table:
            self.bq_table = bq_table
            if self.name is not None and bq_table.table_id != self.name:
                raise ValueError(
                    f"provided bq_table does not match name, {bq_table.table_id} != {self.name}"
                )
        else:
            self.bq_table = get_table(client, project_id, dataset_id, self.name)
        self.full_table_name: str = self.bq_table.full_table_id.replace(":", ".")
        self.buffer: List = []
        self.total_written: int = 0
        self.total_errors: int = 0
        self.bad_records: List = []

    def size(self) -> int:
        since_beginning_of_time = self._partition_field_clause()
        result = self.client.query(
            f"SELECT COUNT(1) as count FROM `{self.full_table_name}` WHERE {since_beginning_of_time}"
        ).result()
        for row in result:
            return row.values()[0]

    def iter_all(self) -> Iterator[Any]:
        """Returns an iterator over all records in a table since 2000, the beginning of all time."""
        since_beginning_of_time = self._partition_field_clause()
        query = (
            f"SELECT * FROM `{self.full_table_name}` WHERE {since_beginning_of_time}"
        )
        for row in self.client.query(query).result():
            # leveraging bigquery's built-in pagination instead of reimplementing our own.
            yield self.mapper(row)

    def all(self) -> Sequence[Any]:
        """Returns all records in a table since 2000, the beginning of all time."""
        return [r for r in self.iter_all()]

    def write_all(self, records) -> Sequence[BQError]:
        """Immediately write all records."""
        if not records:
            return []
        errors = self.client.insert_rows(
            table=self.bq_table,
            rows=records,
            skip_invalid_rows=False,
            ignore_unknown_values=False,
        )
        return self._filter_errors(errors)

    def buffered_write(self, record: Dict):
        """Buffer write a record.

        Buffer will be flushed when size is >= DEFAULT_BUFFER_SIZE or when `flush` is called.
        """
        if record:
            self.buffer.append(record)
        if len(self.buffer) >= self.DEFAULT_BUFFER_SIZE:
            self._write_buffer()

    def flush(self):
        """Flushes the write buffer."""
        self._write_buffer()

    def mapper(self, row):
        """(Abstract) Maps a BQ row to a BanditML model."""
        raise NotImplementedError()

    def status(self):
        s = (
            f"wrote {self.total_written} "
            f"with {self.total_errors} errors "
            f"to {self.name}."
        )
        if self.total_errors > 0:
            s += f"\nsample of bad records:\n{self.bad_records[-5:]}"
        return s

    @staticmethod
    def _filter_errors(errors):
        filtered_errors = []
        for e1 in errors:
            for e2 in e1["errors"]:
                if e2.get("reason", None) != "stopped":
                    filtered_errors.append(e1)
        return filtered_errors

    def _partition_field_clause(self):
        return (
            f"""DATE({self.partition_field}) > "2000-01-01" """
            if self.partition_field
            else ""
        )

    def _write_buffer(self):
        num_records = len(self.buffer)
        errors = self.write_all(self.buffer)
        self.bad_records.extend(
            {"record": self.buffer[error["index"]], "errors": error["errors"]}
            for error in errors
        )
        self.buffer = []
        num_errors = len(errors)
        num_successes = num_records - num_errors
        self.total_written += num_successes
        self.total_errors += num_errors


def get_table(client, project_id, dataset_id, name) -> bigquery.Table:
    fulltable = _fulltable(project_id, dataset_id, name)
    try:
        return client.get_table(fulltable)
    except google.api_core.exceptions.NotFound:
        raise TableNotFound(fulltable)


class TableNotFound(Exception):
    def __init__(self, fulltable):
        self.fulltable = fulltable

    def __str__(self):
        return f"Table {self.fulltable} does not exist"


def _fulltable(project_id, dataset_id, table):
    return f"{project_id}.{dataset_id}.{table}"



from typing import Tuple

import torch

from . import model_constructors, utils

logger = utils.get_logger(__name__)


def calculate_feature_importance(reward_type, feature_names, X, y):
    """Uses random forest to calculate feature importance."""

    X = X["X_float"]

    model = model_constructors.build_random_forest(reward_type)
    model.fit(X, y.squeeze())

    zipped_importances = zip(model.feature_importances_, feature_names)
    return sorted(zipped_importances, reverse=True)


def display_feature_importances(scores: Tuple[float, str]):
    for score, feature_name in scores:
        logger.info(f"{feature_name}: {score}")


def keep_top_n_features(n, X, feature_order, feature_scores):
    """Keep top n features based on feature importance."""

    top_n_feature_names = []
    top_n_feature_idxs = []
    for score, feature_name in feature_scores[:n]:
        top_n_feature_names.append(feature_name)
        top_n_feature_idxs.append(feature_order.index(feature_name))

    idxs = torch.tensor(top_n_feature_idxs)
    X_float = torch.index_select(X["X_float"], 1, idxs)
    X["X_float"] = X_float
    return X, top_n_feature_names



## How to release package to PyPi:
## 1) cd in package directory banditml/
## 2) bump version in setup.py
## 3) pipenv run python setup.py sdist bdist_wheel
## 4) pipenv run twine upload dist/* --skip-existing
## 5) See new version at https://pypi.org/project/banditml/


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="banditml",
    version="1.0.2",
    author="Edoardo Conti, Lionel Vital, Joseph Gilley",
    author_email="team@banditml.com",
    description="Lightweight library for training & serving contextual bandit models.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/banditml/banditml",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=[
        "torch~=1.7.0",
        "pandas~=1.1.4",
        "google-cloud-bigquery~=1.24",
        "scikit-learn>=0.22.2.post1,<0.23",
        "skorch~=0.9.0",
    ],
)



from typing import Dict, Optional

import pandas as pd
from google.cloud import bigquery
from google.oauth2 import service_account

from ..utils.utils import get_logger

logger = get_logger(__name__)


class BigQueryReader:
    def __init__(self, credential_path: str):
        credentials = service_account.Credentials.from_service_account_file(
            filename=credential_path,
            scopes=["https://www.googleapis.com/auth/cloud-platform"],
        )
        project_id = credentials.project_id
        self.client = bigquery.Client(credentials=credentials, project=project_id)

    def get_dataset(self, query: str) -> pd.DataFrame:
        """Execute adhoc / one-off blocking query."""
        return self.client.query(query).result().to_dataframe()
